use std::borrow::BorrowMut;
use std::cell::Ref;
use std::cell::RefCell;
use std::ops::DerefMut;
use std::sync::Arc;

use diesel::Connection;
use diesel::QueryResult;
use diesel::prelude::*;
use crate::conf;
use crate::conf::CONFIG;
use crate::data::PageError;
use crate::data::{self, Page, PageManager};
use crate::schema::pages;

#[derive(diesel::MultiConnection)]
enum AnyConnection {
    Postgresql(PgConnection),
    Sqlite(SqliteConnection),
}

pub struct GitPage {
    conn: Arc<AnyConnection>,
    id: i32,
}

impl GitPage {
    fn get_connection(&self) -> &mut AnyConnection {
    }
}

impl Page for GitPage {
    type Id = i32;

    fn id(&self) -> Self::Id {
        self.id.clone()
    }
    
    fn author(&self) -> String {
        use crate::schema::pages::dsl::*;
        
        let connection = self.get_connection();
        
        let page: Vec<String> = pages
            .filter(id.eq(self.id))
            .limit(1)
            .select(author)
            .load(connection)
            .expect("Failed to load page");

        // Return the author, assuming at least one result
        page.get(0).cloned().unwrap_or_else(|| "Unknown author".to_string())
    }

    fn routing(&self) -> crate::data::PageRouting {
        todo!()
    }

    fn version(&self) -> String {
        use crate::schema::git_page_source::dsl::*;

        let connection = self.get_connection();

        let git_page_source: Vec<String> = crate::schema::git_page_source::table
            .filter(page_id.eq(self.id))
            .limit(1)
            .select(repo_url)
            .load(connection)
            .expect("Failed to load git page source");

        git_page_source.get(0).cloned().unwrap_or_else(|| "Unknown version".to_string())
    }

    fn update(&mut self) -> Result<(), crate::data::PageError> {
        todo!()
    }
}

pub struct GitPageManager<'a> {
    pub connection: RefCell<AnyConnection>,
    pages: Vec<GitPage<'a>>,
}

impl GitPageManager<'_> {
    pub fn new() -> Self {
        let connection = AnyConnection::establish("sqlite://git.db").unwrap();
        GitPageManager {
            connection: RefCell::new(connection),
            pages: Vec::new(),
        }
    }
}


// Filesystem-based storage
// - D [PAGES_TEMP]
//   - D [git_url]
// - D [PAGES_ROOT]
//   - D [UUID]
//     - F? CNAME
//     - F? All other files
//     - F? .git
impl<'a> GitPageManager<'a> {
    fn create(&mut self, url: String, owner: String) -> Result<&GitPage<'a>, PageError> {
        // - Prepare data
        // - Insert page into database (STATUS: FETCHING)
        // - Clone repo to local storage (temp)
        // - Gather CNAME and other data from repo
        // - Validate data (Perform checks)
        // - Update page in database (STATUS: APPLYING)
        // - Move repo to final storage location
        // - Validate
        // - Update page in database (STATUS: READY)
        
        // Insert the page into the database
        let connection = &mut *self.connection.borrow_mut();
        let page_id: i32 = 0;
        {
            use crate::schema::pages::dsl::*;
            let new_page = crate::models::NewPage {
                name: url.clone(),
                author: owner.clone(),
                status: 1,
            };

            let res_insert = diesel::insert_into(pages)
                .values(&new_page)
                .execute(connection);

            if res_insert.is_err() {
                return Err(PageError::InternalServerError(res_insert.err().unwrap().to_string()));
            }

            // Get the ID of the page, given name and author
            let res_page_id: QueryResult<i32> = pages
                .filter(name.eq(url.clone()))
                .filter(author.eq(owner.clone()))
                .select(id)
                .first(connection);
        }
        {
            // Git tracking attachment
            use crate::schema::git_page_source::dsl::*;
            let res_insert_git = diesel::insert_into(git_page_source)
                .values(&crate::models::NewGitPageSource {
                    page_id: 0,
                    repo_url: url.clone(),
                    branch: "master".to_string(),
                })
                .execute(connection);
        }


        // Clone to filesystem (temp)
        let temp_path = format!("{}/{}", CONFIG.temp_dir, url);
        let _ = std::fs::create_dir_all(&temp_path);
        let res_git_repo = git2::Repository::clone(&url, &temp_path);

        if res_git_repo.is_err() {
            return Err(PageError::FetchError(res_git_repo.err().unwrap().to_string()));
        }

        // Open CNAME file
        let cname_path = format!("{}/CNAME", temp_path);
        let cname = std::fs::read_to_string(&cname_path);

        // CNAME follows GitHub's rules
        // Each line is a domain
        let domains: Vec<String> = cname.unwrap_or("".to_string())
            .split("\n")
            .map(|s| s.to_string())
            .collect();

        // TODO: Validation step here - Put this off for now

        // Move to final storage location
        let final_path = format!("{}/{}", CONFIG.pages_root, page_id);
        let _ = std::fs::create_dir_all(&final_path);
        let _ = std::fs::rename(&temp_path, &final_path);

        // Update the page in the database (we are now ready)
        {
            use crate::schema::pages::dsl::*;
            let res_update = diesel::update(pages)
                .filter(id.eq(page_id))
                .set(status.eq(3))
                .execute(connection);

            if res_update.is_err() {
                return Err(PageError::InternalServerError(res_update.err().unwrap().to_string()));
            }
        }

        // Create the page object
        let page = GitPage {
            conn: 
            id: page_id,
        };

        // Add the page to the list of pages
        self.pages.push(page);

        // Return the page
        let res_page = self.page(page_id);

        match res_page {
            Ok(page) => Ok(page),
            Err(err) => Err(PageError::InternalServerError("Failed to create page".to_string())),
        }

    }
}

impl<'a> PageManager<'a, GitPage<'a>> for GitPageManager<'a> {
    
    fn query(&self, author: String, path: String, host: Option<String>) -> Result<crate::data::PageQueryResult, crate::data::PageError> {
        todo!()
    }

    fn page(&self, id: <GitPage<'_> as data::Page>::Id) -> Result<&GitPage<'a>, crate::data::PageError> {
        let page = self.pages.iter().find(|page| page.id() == id);
        match page {
            Some(page) => Ok(page),
            None => Err(crate::data::PageError::NotFound(format!("Page not found: {}", id))),
        }
    }

    fn page_mut(&mut self, id: <GitPage<'_> as data::Page>::Id) -> Result<&mut GitPage<'a>, crate::data::PageError> {
        let page = self.pages.iter_mut().find(|page| page.id() == id);
        match page {
            Some(page) => Ok(page),
            None => Err(crate::data::PageError::NotFound(format!("Page not found: {}", id))),
        }
    }

    fn pages_iter(&'a self) -> impl Iterator<Item = &'a GitPage<'a>> {
        self.pages.iter()
    }    
}