/// Data storage and retrieval

/* -------------------------------------------------------------------------- */
/*                                    Data                                    */
/* -------------------------------------------------------------------------- */

use serde::{Deserialize, Serialize};

pub enum PageError {
    NotFound(String),
    Unauthorized(String),
    InternalServerError(String),
    Corrupted(String),
    FetchError(String),
}

pub trait Page: Sized {
    type Id;

    /// Gets the identifier of the page
    /// This can be used to differentiate between pages
    /// 
    /// # Returns
    /// 
    fn id(&self) -> Self::Id;
    fn author(&self) -> String;
    fn routing(&self) -> PageRouting;
    /// Get the revision of the page
    /// 
    /// # Returns
    /// 
    /// * `String` - The revision of the page - This is implementation defined
    fn version(&self) -> String;
    fn update(&mut self) -> Result<(), PageError>;
}

#[derive(Serialize, Deserialize)]
pub struct PageRouting {
    pub hosts: Option<Vec<String>>,
}

pub enum PageStatus {
    // 200
    Ok = 200,
    Created = 201,
    Accepted = 202,
    NoContent = 204,
    ResetContent = 205,

    // 300

    // 400
    BadRequest = 400,
    Unauthorized = 401,
    // No one cares about 402...
    Forbidden = 403,
    NotFound = 404
}

pub struct PageQueryResult {
    pub content: Option<String>,
    pub status: PageStatus
}

pub trait PageManager<'a, P: Page + 'a> {
    /// Resolve a page
    /// 
    /// # Arguments
    /// 
    /// * `author` - The author of the repository - This is required
    /// * `host` - The host of the repository - This allows CNAMEs to be used to point to the repository
    /// * `path` - The query path - This is treated either as a file path (in the case of a user page),
    ///    or as a repository path (in the case of a project page)
    /// 
    /// # Returns
    /// 
    /// * `PageQueryResult` - The result of the query - This can be page contents, etc
    /// * `PageError` - Any error that occurred processing the query
    /// 
    /// Note that page errors, such as Error 404, are not an error that occurs processing the query
    /// Thus, they are returned via the `PageQueryResult` instead of the `PageError`
    fn query(&self, author: String, path: String, host: Option<String>) -> Result<PageQueryResult, PageError>;

    fn page(&self, id: P::Id) -> Result<&P, PageError>;
    fn page_mut(&mut self, id: P::Id) -> Result<&mut P, PageError>;
    fn pages_iter(&'a self) -> impl Iterator<Item = &'a P>;
    /// Delete the page
    /// This will remove the page from being hosted or stored
    /// 
    /// # Returns
    /// 
    /// * `Result<(), (PageError, Self)>` - If the page was deleted successfully, this will return `Ok(())`
    ///   Otherwise, it will return `Err((PageError, Self))` with the error and the page
    /// 
    /// Consequently, since the page is not returned if it is deleted successfully, the page is consumed
    /// And since it returns it if it fails, it is recoverable for failure handling
    fn delete(&mut self, id: P::Id) -> Result<(), PageError>;
}